#pragma once
#include "PDTExp.h"
#include "PDTVar.h"
#include "PDTVecConst.h"
#include "PDTNumConst.h"

class PDTSequenceAccessNum :  public PDTExpNum, public PDTVar, public TraverseIfc <PDTSequenceAccessNum, PDTNode> { 
public: 
 TRAVERSAL_IFC_IMPL(PDTSequenceAccessNum)
public:
    PDTSequenceAccessNum(PDTCodeGraph & CG, PDTSequence & left, PDTExpNum & right) :
        PDTExpNum(CG), PDTVar(left.name) {
            addChild(&left);
            addChild(&right);

            left.addParent(this);
            right.addParent(this);
        }

    NodeTypeEnum getType() {return PDTSequenceAccessNum_Em;}
    
    virtual bool isSequenceAccessNum() {
    	return true;
    }

    virtual PDTNode* getNode() {
        return this;
    }

    PDTSequence * getLeft() const {
        return dynamic_cast<PDTSequence*> (children[0]);
    }
    PDTExpNum* getRight() const {
        return dynamic_cast<PDTExpNum*> (children[1]);
    }
    
    PDTVar * getIndexVar() {
    	if (children[1]->isVar())
    		return dynamic_cast<PDTVar*>(children[1]);
        return NULL;
    }

    virtual bool eval(PDTCodeGraph &CG)
    {
    	getRight()->eval(CG); //compute index
    	PDTNode * idxVal = getRight()->getNValue();
    	int index = (dynamic_cast<PDTNumConstInt*>(idxVal))->getValue(); //int index
   	 	PDTNode * element = getLeft()->getElement(index); // get Value at index
   	 	if (getLeft()->isElementInt()) //If Int, initialize nValue to int
		{
   	 		int intElement = (dynamic_cast<PDTNumConstInt*>(element))->getValue();
   	 		nValue = new PDTNumConstInt(CG, intElement);
			return true;
		}

		if (getLeft()->isElementBoolean())
		{
			int boolElement = (dynamic_cast<PDTBoolConst*>(element))->getValue();
			nValue = new PDTBoolConst(CG, boolElement);
			return true;
		}

   	 	if ( nValue == NULL ){
   	 		cerr << "Warning: array access " << getLeft()->name << "[" << index << " ]" << " is not available." << endl;
            return false;
   	 	}
        return true;
   	}


    virtual stringstream & toString(stringstream & ss) const {
        getLeft()->toString(ss);
        ss << "at (";
        getRight()->toString(ss);
        ss << ")";
        return ss;
    }
};

class PDTSequenceAccessBool :  public PDTExpBool, public PDTVar, public TraverseIfc <PDTSequenceAccessBool, PDTNode> { 
public: 
 TRAVERSAL_IFC_IMPL(PDTSequenceAccessBool)
public:
    PDTSequenceAccessBool(PDTCodeGraph & CG, PDTSequence & left, PDTExpNum & right) :
        PDTExpBool(CG), PDTVar(left.name) {
            addChild(&left);
            addChild(&right);

            left.addParent(this);
            right.addParent(this);
        }

    NodeTypeEnum getType() {return PDTSequenceAccessBool_Em;}

    virtual bool isSequenceAccessBool() {
    	return true;
    }

    virtual PDTNode* getNode() {
        return this;
    }

    PDTSequence * getLeft() const {
        return dynamic_cast<PDTSequence*> (children[0]);
    }
    PDTExpNum * getRight() const {
        return dynamic_cast<PDTExpNum*> (children[1]);
    }

    virtual bool eval(PDTCodeGraph &CG)
    {
    	getRight()->eval(CG); //compute index
    	PDTNode * idxVal = getRight()->getNValue();
    	int index = (dynamic_cast<PDTNumConstInt*>(idxVal))->getValue(); //int index
   	 	PDTNode * element = getLeft()->getElement(index); // get Value at index
   	 	if (getLeft()->isElementInt()) //If Int, initialize nValue to int
		{
   	 		int intElement = (dynamic_cast<PDTNumConstInt*>(element))->getValue();
   	 		nValue = new PDTNumConstInt(CG, intElement);
			return true;
		}

		if (getLeft()->isElementBoolean())
		{
			int boolElement = (dynamic_cast<PDTBoolConst*>(element))->getValue();
			nValue = new PDTBoolConst(CG, boolElement);
			return true;
		}

   	 	if ( nValue == NULL ){
   	 		cerr << "Warning: array access " << getLeft()->name << "[" << index << " ]" << " is not available." << endl;
   	 	}
        return true;
   	}

    virtual stringstream & toString(stringstream & ss) const {
        getLeft()->toString(ss);
        ss << "at (";
        getRight()->toString(ss);
        ss << ")";
        return ss;
    }
};
